"""Get a variable from a passed-in ``prop``."""

from tdom import html


def Hello(name):
    """A simple hello component."""
    return html(t"<div>Hello {name}</div>")


def main():
    """Main entry point."""
    result = Hello(name="tdom")
    return result



"""Shorthand syntax for attribute values means no need for double quotes."""

from tdom import html


def main():
    """Main entry point."""
    result = html(t'<div class={"Container1".lower()}>Hello World</div>')
    return result



"""Use ``children`` as a built-in "prop"."""

from tdom import html


def Heading(title):
    """Default heading."""
    return html(t"<h1>{title}</h1>")


def main():
    """Main entry point."""
    result = html(t'<{Heading} title="My Title" />')
    return result



"""An expression which chooses subcomponent based on condition."""

from tdom import html


def DefaultHeading():
    """The default heading."""
    return html(t"<h1>Default Heading</h1>")


def OtherHeading():
    """Another heading used in another condition."""
    return html(t"<h1>Other Heading</h1>")


def Body(heading):
    """Render the body with a heading based on which is passed in."""
    return html(t"<body>{heading if heading else DefaultHeading}</body>")


def main():
    """Main entry point."""
    result = html(t"<{Body} heading={OtherHeading}/>")
    return result



"""Test an example."""
from . import main
from tdom import Element


def test_main() -> None:
    """Ensure the demo matches expected."""
    result: Element = main()
    assert isinstance(result, Element)
    assert result.name == 'div'
    assert result.props == {"class": "container"}
    assert result.children == [{'data': 'Hello World', 'type': 3}]



"""Cover the examples in Andrea's demo."""

from tdom import html, svg, unsafe


def test_automatic_quotes():
    """Automatic quotes with safe escapes."""
    result = html(t"""<p class={'test & "test"'}>Hello</p>""")
    assert str(result) == '<p class="test &amp; &quot;test&quot;">Hello</p>'


def test_sanitized_content():
    """Sanitized content out of the box"""
    from random import random

    result = html(t"Some random number: {random()}")
    # QUESTION: Not sure what to test here
    assert str(result).startswith("Some random number")


def test_self_closing_tags():
    """Sanitized content out of the box"""
    value = "Hello"
    result = html(t"<textarea placeholder={value} />")
    assert str(result) == '<textarea placeholder="Hello"></textarea>'


def test_special_attributes():
    """Data and ARIA"""
    data = {"a": 1, "b": 2}
    aria = {"role": "button", "label": "Click me"}
    result = html(t"<div data={data} aria={aria} />")
    assert (
        str(result)
        == '<div data-a="1" data-b="2" role="button" aria-label="Click me"></div>'
    )


def test_click_handler():
    """Bind a Python function to an element event handler."""

    def on_click(event):
        import js

        js.alert(event.type)

    result = html(t"<div @click={on_click} />")
    assert str(result) == '<div onclick="self.python_listeners?.[0](event)"></div>'


def test_ignore_voided():
    """Voided elements."""
    result = html(t"<hr />")
    assert str(result) == "<hr>"


def test_void_siblings():
    """Voided elements' siblings."""
    result = html(t"<p /><hr><p />")
    assert str(result) == "<p></p><hr><p></p>"


def test_dev_comments():
    """Developer comments."""
    result = html(t"<!--#1--><!--#2#--><!--##--><!--3#-->")
    assert str(result) == "<!--#1--><!--3#-->"


def test_svg():
    """preseved XML/SVG self closing nature."""
    result = html(
        t"""
    <svg>
      {svg(t'<rect width="200" height="100" rx="20" ry="20" fill="blue" />')}
    </svg>
    """
    )
    assert str(result) == (
        '<svg><rect width="200" height="100" rx="20" ry="20" fill="blue" /></svg>'
    )


def test_style():
    """Style attribute."""
    style = {"color": "red", "font-size": "12px"}
    result = html(t"<div style={style} />")
    assert str(result) == '<div style="color:red;font-size:12px"></div>'


def test_unsafe():
    """Unsafe strings."""
    result = html(t"<div>{unsafe('<hr>')}</div>")
    assert str(result) == '<div><hr></div>'


def test_component():
    """Render a t-string that references a component."""

    def Component(a:str, b:int, children:list):
        return html(
            t"""
                <div a={a} b={b}>
                    {children}
                </div>
            """
        )

    result = html(
        t"""
            <{Component} a="1" b={2}>
                <p>Hello Components!</p>
            <//>
        """
    )

    assert "<p>Hello Components!</p>" in str(result)


def test_component_without_children():
    """Render a t-string that references a component."""

    def Component(a:str, b:int):
        return html(
            t"""
                <div a={a} b={b} />
            """
        )

    result = html(
        t"""
            <{Component} a="1" b={2} />
        """
    )

    assert '<div a="1" b="2"></div>' in str(result)


def test_lists_within_layout():
    """A template in a template."""

    names = ["John", "Jane", "Jill"]
    result = html(
        t"""
            <ul>
                {[html(t"<li>{name}</li>") for name in names]}
            </ul>
        """
    )

    assert "<li>John</li>" in str(result)
    assert "<li>Jane</li>" in str(result)
    assert "<li>Jill</li>" in str(result)



"""Children as props."""

from tdom import html


def Heading(title, children):
    """The default heading."""
    return html(t"<h1>{title}</h1><div>{children}</div>")


def main():
    """Main entry point."""
    result = html(t'<{Heading} title="My Title">Child<//>')
    return result



"""Simple function component, nothing dynamic, that returns a VDOM."""
from tdom import html


def Heading():
    """The default heading."""
    return html(t"<h1>My Title</h1>")


def main():
    """Main entry point."""
    result = html(t"<{Heading} />")
    return result



"""Generators as components."""

from tdom import html


def Todos():
    """A sequence of li items."""
    for todo in ["First", "Second"]:  # noqa B007
        # TODO Andrea need to add generator support
        yield html(t"<li>{todo}</li>")


def main():
    """Main entry point."""
    result = html(t"<ul><{Todos}/></ul>")
    return result



"""Generate a list of VDOMs then use in a render."""

from tdom import html


def main():
    """Main entry point."""
    message = "Hello"
    names = ["World", "Universe"]
    items = [html(t"<li>{label}</li>") for label in names]
    result = html(
        t"""
            <ul title="{message}">
              {items}
            </ul>
            """
    )
    return result



"""A variation of the to do list, possibly duplicated in another example."""
from tdom import html
from viewdom import render


def Todo(label):
    """An individual to do component."""
    return html(t"<li>{label}</li>")


def TodoList(todos):
    """A to do list component."""
    return html(t"<ul>{[Todo(label) for label in todos]}</ul>")


def TodoApp(title, todolist):
    """A to do list application."""
    return html(t"<h1>{title}</h1>{todolist}")


def main():
    """Main entry point."""
    todos = ["first"]
    todo_list = TodoList(todos)
    return render(
        html(
            """
      <{TodoApp} title="My Todos" todolist={todo_list} />
    """
        )
    )


result = main()
expected = "<h1>My Todos</h1><ul><li>first</li></ul>"


def main():
    """Main entry point."""
    return result



"""
Port the code from viewdom.examples.expressions.
"""

import pytest

from tdom import html


@pytest.fixture(scope="module")
def fixture_name():
    return "Fixture Name"

def no_fixture_name():
    return "Fixture Name"


def test_insert_variable():
    """Template is in a function, and `name` comes from that scope."""
    name = "World"
    result = html(t"<div>Hello {name}</div>")
    assert str(result) == "<div>Hello World</div>"


def test_from_import():
    """A symbol is imported from another module."""
    name = pytest.__name__
    result = html(t"<div>Hello {name}</div>")
    assert str(result) == "<div>Hello pytest</div>"


def test_from_function_arg(fixture_name=no_fixture_name):
    """A symbol is passed into a function."""
    result = html(t"<div>Hello {fixture_name}</div>")
    assert str(result) == "<div>Hello Fixture Name</div>"


def test_python_expression():
    """Use an expression which adds two numbers together."""
    result = html(t"<div>Hello {2+2}</div>")
    assert str(result) == "<div>Hello 4</div>"


def test_call_function():
    """Use an expression which calls a function."""

    def make_bigly(name: str) -> str:
        """A function returning a string, rather than a component."""
        return f"BIGLY: {name.upper()}"

    name = "World"
    result = html(t"<div>Hello {make_bigly(name)}</div>")
    assert str(result) == "<div>Hello BIGLY: WORLD</div>"



"""Loop through values in a template and render them in a nested template."""

from tdom import html


def main():
    """Main entry point."""
    message = "Hello"
    names = ["World", "Universe"]
    result = html(
        t"""
            <ul title="{message}">
                {[
                    html(t'<li>{name}</li>')
                    for name in names
                ]}
            </ul>
            """
    )
    return result



"""Components can be any kind of dataclass_component."""

from dataclasses import dataclass

from tdom import html


@dataclass
class Greeting:
    """Give a greeting."""

    name: str

    def __call__(self):
        """Render to a string."""
        return f"Hello {self.name}"


def main():
    """Render a template to a string."""
    greeting = Greeting(name="tdom")
    # TODO Teach the constructor to make dataclass components
    result = html(t"<div><{greeting} /></div>")
    return result



"""Simple example of inserting a variable value into a template."""

from tdom import html


def main():
    """Main entry point."""
    name = "tdom"
    result = html(t"<div>Hello {name}</div>")
    return result



"""Overriding a default "built-in" component."""

from tdom import html


def DefaultHeading():  # pragma: nocover
    """The default heading."""
    return html(t"<h1>Default Heading</h1>")


def OtherHeading():
    """Another heading used in another condition."""
    return html(t"<h1>Other Heading</h1>")


def Body(heading):
    """Render the body with a heading based on which is passed in."""
    return html(t"<body><{heading} /></body>")


def main():
    """Main entry point."""
    result = html(t"<{Body} heading={OtherHeading}/>")
    return result



"""
Port the code from viewdom.examples.variables.
"""

import pytest

from tdom import html


@pytest.fixture(scope="module")
def fixture_name():
    return "Fixture Name"

def no_fixture_name():
    return "Fixture Name"

def test_insert_variable():
    """Template is in a function, and `name` comes from that scope."""
    name = "World"
    result = html(t"<div>Hello {name}</div>")
    assert str(result) == "<div>Hello World</div>"


def test_from_import():
    """A symbol is imported from another module."""
    name = pytest.__name__
    result = html(t"<div>Hello {name}</div>")
    assert str(result) == "<div>Hello pytest</div>"


def test_from_function_arg(fixture_name=no_fixture_name):
    """A symbol is passed into a function."""
    result = html(t"<div>Hello {fixture_name}</div>")
    assert str(result) == "<div>Hello Fixture Name</div>"



import inspect
from types import GeneratorType
from .parser import _instrument, _prefix
from .dom import Fragment, Node, Text
from .dom import COMMENT, ELEMENT, FRAGMENT
from .dom import _IS_MICRO_PYTHON, _appendChildren, _replaceWith, parse as domify

if not _IS_MICRO_PYTHON:
  import inspect


def _as_comment(node):
  return lambda value: _replaceWith(node, _as_node(value))


def _as_component(node, components):
  def component(value):
    def reveal():
      if _IS_MICRO_PYTHON or ('children' in inspect.signature(value).parameters):
        props = { 'children': node['children'] }
        for k, v in node['props'].items():
          props[k] = v
      else:
        props = node['props'].copy()

      if _IS_MICRO_PYTHON:
        try:  
          result = value(**props)
        except TypeError as e:
          if "unexpected keyword argument" in str(e):
            del props['children']
            result = value(**props)
          else:
            raise e
      else:
        result = value(**props)

      _replaceWith(node, _as_node(result))

    components.append(reveal)

  return component


def _as_node(value):
  if isinstance(value, Node):
    return value
  if isinstance(value, (list, tuple, GeneratorType)):
    node = Fragment()
    _appendChildren(node, value)
    return node
  if callable(value):
    # TODO: this could be a hook pleace for asyncio
    #       and run to completion before continuing
    return _as_node(value())
  return Text(value)


def _as_prop(node, name, listeners):
  props = node['props']

  def aria(value):
    for k, v in value.items():
      lower = k.lower()
      props[lower if lower == 'role' else f'aria-{lower}'] = v

  def attribute(value):
    props[name] = value

  def dataset(value):
    for k, v in value.items():
      props[f'data-{k.replace('_', '-')}'] = v

  def listener(value):
    if value in listeners:
      i = listeners.index(value)
    else:
      i = len(listeners)
      listeners.append(value)

    props[name] = f'self.python_listeners?.[{i}](event)'

  def style(value):
    if isinstance(value, dict):
      props['style'] = ';'.join([f'{k}:{v}' for k,v in value.items()])
    else:
      props['style'] = value

  if name[0] == '@':
    name = 'on' + name[1:].lower()
    return listener
  if name == 'style':
    return style
  if name == 'aria':
    return aria
  # TODO: find which other node has a `data` attribute
  elif name == 'data' and node['name'].lower() != 'object':
    return dataset
  else:
    return attribute


def _set_updates(node, updates, path):
  type = node['type']
  if type == ELEMENT:
    if node['name'] == _prefix:
      updates.append(_Update(path, _Component()))

    remove = []
    props = node['props']
    for key, name in props.items():
      if key.startswith(_prefix):
        remove.append(key)
        updates.append(_Update(path, _Attribute(name)))

    for key in remove:
      del props[key]

  if type == ELEMENT or type == FRAGMENT:
    i = 0
    for child in node['children']:
      _set_updates(child, updates, path + [i])
      i += 1

  elif type == COMMENT and node['data'] == _prefix:
    updates.append(_Update(path, _Comment()))



class _Attribute:
  def __init__(self, name):
    self.name = name

  def __call__(self, node, listeners):
    return _as_prop(node, self.name, listeners)


class _Comment:
  def __call__(self, node):
    return _as_comment(node)


class _Component:
  def __call__(self, node, updates):
    return _as_component(node, updates)


class _Update:
  def __init__(self, path, update):
    self.path = path
    self.value = update



def _parse(template, length, svg):
  updates = []
  content = _instrument(template, svg)
  fragment = domify(content, svg)

  if len(fragment['children']) == 1:
    node = fragment['children'][0]
    if node['type'] != ELEMENT or node['name'] != _prefix:
      fragment = node

  _set_updates(fragment, updates, [])

  if len(updates) != length:
    raise ValueError(f'{len(updates)} updates found, expected {length}')

  return [fragment, updates]



"""A variation of the to do list, possibly duplicated in another example."""
from dataclasses import dataclass
from typing import Iterable

from tdom import html
from viewdom import render
from viewdom import VDOM


def Todo(label):
    """An individual to do component."""
    return html(t"<li>{label}</li>")


@dataclass
class TodoList:
    """A class-based to do list with passed-in data."""

    todos: Iterable

    def __call__(self) -> VDOM:
        """Render to a string."""
        return html(t"<ul>{[Todo(label) for label in self.todos]}</ul>")


def TodoApp(title, todolist):
    """A to do list application."""
    return html(t"<h1>{title}</h1>{todolist}")


def main():
    """Main entry point."""
    todos = ["first"]
    tl = TodoList(todos)
    return render(
        html(
            """
      <{TodoApp} title="My Todos" todolist={tl} />
    """
        )
    )


result = main()
expected = "<h1>My Todos</h1><ul><li>first</li></ul>"


def main():
    """Main entry point."""
    return result



"""Use a nested f-string to communicate a dynamic attribute value."""

from tdom import html


def main():
    """Main entry point."""
    result = html(t'<div class="{f"container{1}"}">Hello World</div>')
    return result



"""The prop has a default value, so caller doesn't have to provide it."""

from tdom import html


def Hello(name="World"):
    """A simple hello component."""
    return html(t"<div>Hello {name}</div>")


def main():
    """Main entry point."""
    result = Hello()
    return result



"""Prop values from scope variables."""
from tdom import html


def Heading(title):
    """The default heading."""

    return html(t"<h1>{title}</h1>")


this_title = "My Title"


def main():
    """Main entry point."""
    result = html(t"<{Heading} title={this_title} />")
    return result



"""Optional props."""
from tdom import html


def Heading(title="My Title"):
    """The default heading."""

    return html(t"<h1>{title}</h1>")


def main():
    """Main entry point."""
    result = html(t"<{Heading} />")
    return result



"""Pass a Python symbol as part of an expression."""

from tdom import html


def Heading(title):
    """The default heading."""
    return html(t"<h1>{title}</h1>")


def main():
    """Main entry point."""
    result = html(t'<{Heading} title={"My Title"} />')
    return result
